@@ -612,7 +612,6 @@ gdk_event_queue_handle_scroll_compression (GdkDisplay *display)
   GList *l;
   GdkSurface *surface = NULL;
   GdkDevice *device = NULL;
-  GdkEvent *last_event = NULL;
   GList *scrolls = NULL;
   GArray *history = NULL;
   GdkScrollUnit scroll_unit = GDK_SCROLL_UNIT_WHEEL;
@@ -645,9 +644,6 @@ gdk_event_queue_handle_scroll_compression (GdkDisplay *display)
           scroll_unit != scroll_event->unit)
         break;
 
-      if (!last_event)
-        last_event = event;
-
       surface = event->surface;
       device = event->device;
       scroll_unit = scroll_event->unit;
@@ -1178,9 +1174,16 @@ gdk_event_get_axes (GdkEvent  *event,
                     double   **axes,
                     guint     *n_axes)
 {
+  gboolean ret;
+
   g_return_val_if_fail (GDK_IS_EVENT (event), FALSE);
 
-  return GDK_EVENT_GET_CLASS (event)->get_axes (event, axes, n_axes);
+  ret = GDK_EVENT_GET_CLASS (event)->get_axes (event, axes, n_axes);
+
+  if (*axes == NULL)
+    return FALSE;
+
+  return ret;
 }
 
 double *
@@ -1330,7 +1333,7 @@ gdk_event_get_display (GdkEvent *event)
  * gdk_event_get_event_sequence:
  * @event: a `GdkEvent`
  *
- * Retuns the event sequence to which the event belongs.
+ * Returns the event sequence to which the event belongs.
  *
  * Related touch events are connected in a sequence. Other
  * events typically don't have event sequence information.
@@ -1368,6 +1371,8 @@ gdk_event_get_modifier_state (GdkEvent *event)
  * @y: (out): location to put event surface y coordinate
  *
  * Extract the event surface relative x/y coordinates from an event.
+ *
+ * This position is in [surface coordinates](coordinates.html).
  */
 gboolean
 gdk_event_get_position (GdkEvent *event,
@@ -1508,7 +1513,6 @@ gdk_button_event_get_button (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkKeyEvent */
 
 /**
@@ -1969,7 +1973,6 @@ gdk_key_event_get_match (GdkEvent        *event,
 }
 
 /* }}} */
-
 /* {{{ GdkTouchEvent */
 
 /**
@@ -2106,7 +2109,6 @@ gdk_touch_event_get_emulating_pointer (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkCrossingEvent */
 
 /**
@@ -2250,7 +2252,6 @@ gdk_crossing_event_get_detail (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkDeleteEvent */
 
 /**
@@ -2281,7 +2282,6 @@ gdk_delete_event_new (GdkSurface *surface)
 }
 
 /* }}} */
-
 /* {{{ GdkFocusEvent */
 
 /**
@@ -2338,7 +2338,6 @@ gdk_focus_event_get_in (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkScrollEvent */
 
 /**
@@ -2597,7 +2596,6 @@ gdk_scroll_event_get_unit (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkTouchpadEvent */
 
 /**
@@ -2856,7 +2854,6 @@ gdk_touchpad_event_get_pinch_scale (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkPadEvent */
 
 /**
@@ -3030,7 +3027,6 @@ gdk_pad_event_get_group_mode (GdkEvent *event,
 }
 
 /* }}} */
-
 /* {{{ GdkMotionEvent */
 
 /**
@@ -3192,7 +3188,6 @@ gdk_event_get_history (GdkEvent *event,
 }
 
 /* }}} */
-
 /* {{{ GdkProximityEvent */
 
 /**
@@ -3255,7 +3250,6 @@ gdk_proximity_event_new (GdkEventType   type,
 }
 
 /* }}} */
-
 /* {{{ GdkDNDEvent */
 
 /**
@@ -3361,7 +3355,6 @@ gdk_dnd_event_get_drop (GdkEvent *event)
 }
 
 /* }}} */
-
 /* {{{ GdkGrabBrokenEvent */
 
 /**
@@ -3439,3 +3432,5 @@ gdk_grab_broken_event_get_implicit (GdkEvent *event)
 }
 
 /* }}} */
+
+/* vim:set foldmethod=marker expandtab: */
