@@ -25,7 +25,8 @@
  * multiple frames, and will be used for a long time.
  *
  * There are various ways to create `GdkTexture` objects from a
- * [class@GdkPixbuf.Pixbuf], or a Cairo surface, or other pixel data.
+ * [class@GdkPixbuf.Pixbuf], or from bytes stored in memory, a file, or a
+ * [struct@Gio.Resource].
  *
  * The ownership of the pixel data is transferred to the `GdkTexture`
  * instance; you can only make a copy of it, via [method@Gdk.Texture.download].
@@ -282,6 +283,8 @@ gdk_texture_dispose (GObject *object)
 {
   GdkTexture *self = GDK_TEXTURE (object);
 
+  g_clear_pointer (&self->diff_to_previous, cairo_region_destroy);
+
   gdk_texture_clear_render_data (self);
 
   G_OBJECT_CLASS (gdk_texture_parent_class)->dispose (object);
@@ -342,7 +345,9 @@ gdk_texture_init (GdkTexture *self)
  *
  * Creates a new texture object representing the surface.
  *
- * @surface must be an image surface with format `CAIRO_FORMAT_ARGB32`.
+ * The @surface must be an image surface with format `CAIRO_FORMAT_ARGB32`.
+ *
+ * The newly created texture will acquire a reference on the @surface.
  *
  * Returns: a new `GdkTexture`
  */
@@ -361,7 +366,7 @@ gdk_texture_new_for_surface (cairo_surface_t *surface)
                                       * cairo_image_surface_get_stride (surface),
                                       (GDestroyNotify) cairo_surface_destroy,
                                       cairo_surface_reference (surface));
-  
+
   texture = gdk_memory_texture_new (cairo_image_surface_get_width (surface),
                                     cairo_image_surface_get_height (surface),
                                     GDK_MEMORY_DEFAULT,
@@ -395,7 +400,7 @@ gdk_texture_new_for_pixbuf (GdkPixbuf *pixbuf)
 
   bytes = g_bytes_new_with_free_func (gdk_pixbuf_get_pixels (pixbuf),
                                       gdk_pixbuf_get_height (pixbuf)
-                                      * gdk_pixbuf_get_rowstride (pixbuf),
+                                      * (gsize) gdk_pixbuf_get_rowstride (pixbuf),
                                       g_object_unref,
                                       g_object_ref (pixbuf));
   texture = gdk_memory_texture_new (gdk_pixbuf_get_width (pixbuf),
@@ -671,7 +676,49 @@ gdk_texture_do_download (GdkTexture      *texture,
                          guchar          *data,
                          gsize            stride)
 {
-  GDK_TEXTURE_GET_CLASS (texture)->download (texture, format, data,stride);
+  GDK_TEXTURE_GET_CLASS (texture)->download (texture, format, data, stride);
+}
+
+void
+gdk_texture_diff (GdkTexture     *self,
+                  GdkTexture     *other,
+                  cairo_region_t *region)
+{
+  if (self == other)
+    return;
+
+  if (self->previous_texture == other &&
+      g_atomic_pointer_get (&other->next_texture) == self)
+    {
+      cairo_region_union (region, self->diff_to_previous);
+    }
+  else if (other->previous_texture == self &&
+           g_atomic_pointer_get (&self->next_texture) == other)
+    {
+      cairo_region_union (region, other->diff_to_previous);
+    }
+  else
+    {
+      cairo_region_union_rectangle (region,
+                                    &(cairo_rectangle_int_t) {
+                                      0,
+                                      0,
+                                      MAX (self->width, other->width),
+                                      MAX (self->height, other->height)
+                                    });
+    }
+}
+
+void
+gdk_texture_set_diff (GdkTexture     *self,
+                      GdkTexture     *previous,
+                      cairo_region_t *diff)
+{
+  g_assert (self->diff_to_previous == NULL);
+
+  self->previous_texture = previous;
+  self->diff_to_previous = diff;
+  g_atomic_pointer_set (&previous->next_texture, self);
 }
 
 cairo_surface_t *
@@ -774,7 +821,7 @@ gdk_texture_set_render_data (GdkTexture     *self,
                              GDestroyNotify  notify)
 {
   g_return_val_if_fail (data != NULL, FALSE);
- 
+
   if (self->render_key != NULL)
     return FALSE;
 
