@@ -42,6 +42,7 @@
 #include "a11y/atspi/atspi-component.h"
 
 #include "gtkdebug.h"
+#include "gtkprivate.h"
 #include "gtkeditable.h"
 #include "gtkentryprivate.h"
 #include "gtkroot.h"
@@ -81,6 +82,10 @@
  *   hold the TAB_PANEL role and be the target of the CONTROLS
  *   relation with their corresponding tabs (in the stack
  *   switcher or notebook).
+ *
+ * These are the exceptions implemented by GTK itself, but note that application
+ * developers can customize the accessibility tree by implementing the
+ * [iface@Gtk.Accessible] interface in any way they choose.
  */
 
 struct _GtkAtSpiContext
@@ -137,10 +142,10 @@ collect_states (GtkAtSpiContext    *self,
   accessible = gtk_at_context_get_accessible (ctx);
 
   set_atspi_state (&states, ATSPI_STATE_VISIBLE);
+  set_atspi_state (&states, ATSPI_STATE_SHOWING);
 
-  if (ctx->accessible_role == GTK_ACCESSIBLE_ROLE_WINDOW)
+  if (ctx->accessible_role == GTK_ACCESSIBLE_ROLE_APPLICATION)
     {
-      set_atspi_state (&states, ATSPI_STATE_SHOWING);
       if (gtk_accessible_get_platform_state (accessible, GTK_ACCESSIBLE_PLATFORM_STATE_ACTIVE))
         set_atspi_state (&states, ATSPI_STATE_ACTIVE);
     }
@@ -198,6 +203,7 @@ collect_states (GtkAtSpiContext    *self,
 
   if (gtk_at_context_has_accessible_state (ctx, GTK_ACCESSIBLE_STATE_CHECKED))
     {
+      set_atspi_state (&states, ATSPI_STATE_CHECKABLE);
       value = gtk_at_context_get_accessible_state (ctx, GTK_ACCESSIBLE_STATE_CHECKED);
       switch (gtk_tristate_accessible_value_get (value))
         {
@@ -241,7 +247,7 @@ collect_states (GtkAtSpiContext    *self,
         case GTK_ACCESSIBLE_INVALID_TRUE:
         case GTK_ACCESSIBLE_INVALID_GRAMMAR:
         case GTK_ACCESSIBLE_INVALID_SPELLING:
-          set_atspi_state (&states, ATSPI_STATE_INVALID);
+          set_atspi_state (&states, ATSPI_STATE_INVALID_ENTRY);
           break;
         case GTK_ACCESSIBLE_INVALID_FALSE:
         default:
@@ -277,6 +283,44 @@ collect_states (GtkAtSpiContext    *self,
         }
     }
 
+  if (gtk_at_context_has_accessible_state (ctx, GTK_ACCESSIBLE_STATE_VISITED))
+    {
+      value = gtk_at_context_get_accessible_state (ctx, GTK_ACCESSIBLE_STATE_VISITED);
+      if (value->value_class->type == GTK_ACCESSIBLE_VALUE_TYPE_BOOLEAN)
+        {
+          if (gtk_boolean_accessible_value_get (value))
+            set_atspi_state (&states, ATSPI_STATE_VISITED);
+        }
+    }
+
+  if (gtk_at_context_has_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_REQUIRED))
+    {
+      value = gtk_at_context_get_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_REQUIRED);
+      if (gtk_boolean_accessible_value_get (value))
+        set_atspi_state (&states, ATSPI_STATE_REQUIRED);
+    }
+
+  if (gtk_at_context_has_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_MULTI_SELECTABLE))
+    {
+      value = gtk_at_context_get_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_MULTI_SELECTABLE);
+      if (gtk_boolean_accessible_value_get (value))
+        set_atspi_state (&states, ATSPI_STATE_MULTISELECTABLE);
+    }
+
+  if (gtk_at_context_has_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_HAS_POPUP))
+    {
+      value = gtk_at_context_get_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_HAS_POPUP);
+      if (gtk_boolean_accessible_value_get (value))
+        set_atspi_state (&states, ATSPI_STATE_HAS_POPUP);
+    }
+
+  if (gtk_at_context_has_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_AUTOCOMPLETE))
+    {
+      value = gtk_at_context_get_accessible_property (ctx, GTK_ACCESSIBLE_PROPERTY_AUTOCOMPLETE);
+      if (gtk_autocomplete_accessible_value_get (value) != GTK_ACCESSIBLE_AUTOCOMPLETE_NONE)
+        set_atspi_state (&states, ATSPI_STATE_SUPPORTS_AUTOCOMPLETION);
+    }
+
   g_variant_builder_add (builder, "u", (guint32) (states & 0xffffffff));
   g_variant_builder_add (builder, "u", (guint32) (states >> 32));
 }
@@ -294,11 +338,12 @@ collect_relations (GtkAtSpiContext *self,
     { GTK_ACCESSIBLE_RELATION_LABELLED_BY, ATSPI_RELATION_LABELLED_BY },
     { GTK_ACCESSIBLE_RELATION_CONTROLS, ATSPI_RELATION_CONTROLLER_FOR },
     { GTK_ACCESSIBLE_RELATION_DESCRIBED_BY, ATSPI_RELATION_DESCRIBED_BY },
+    { GTK_ACCESSIBLE_RELATION_DETAILS, ATSPI_RELATION_DETAILS },
+    { GTK_ACCESSIBLE_RELATION_ERROR_MESSAGE, ATSPI_RELATION_ERROR_MESSAGE},
     { GTK_ACCESSIBLE_RELATION_FLOW_TO, ATSPI_RELATION_FLOWS_TO},
   };
   GtkAccessibleValue *value;
   GList *list, *l;
-  GtkATContext *target_ctx;
   int i;
 
   for (i = 0; i < G_N_ELEMENTS (map); i++)
@@ -313,13 +358,16 @@ collect_relations (GtkAtSpiContext *self,
 
       for (l = list; l; l = l->next)
         {
-          target_ctx = gtk_accessible_get_at_context (GTK_ACCESSIBLE (l->data));
+          GtkATContext *target_ctx =
+            gtk_accessible_get_at_context (GTK_ACCESSIBLE (l->data));
 
           /* Realize the ATContext of the target, so we can ask for its ref */
           gtk_at_context_realize (target_ctx);
 
           g_variant_builder_add (&b, "@(so)",
                                  gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (target_ctx)));
+
+          g_object_unref (target_ctx);
         }
 
       g_variant_builder_add (builder, "(ua(so))", map[i].s, &b);
@@ -328,27 +376,44 @@ collect_relations (GtkAtSpiContext *self,
 /* }}} */
 /* {{{ Accessible implementation */
 static int
-get_index_in_parent (GtkWidget *widget)
+get_index_in (GtkAccessible *parent,
+              GtkAccessible *child)
 {
-  GtkWidget *parent = gtk_widget_get_parent (widget);
-  GtkWidget *child;
-  int idx;
-
   if (parent == NULL)
     return -1;
 
-  idx = 0;
-  for (child = gtk_widget_get_first_child (parent);
-       child;
-       child = gtk_widget_get_next_sibling (child))
+  guint res = 0;
+  GtkAccessible *candidate;
+  for (candidate = gtk_accessible_get_first_accessible_child (parent);
+       candidate != NULL;
+       candidate = gtk_accessible_get_next_accessible_sibling (candidate))
     {
-      if (child == widget)
-        return idx;
+      g_object_unref (candidate);
 
-      if (!gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
+      if (candidate == child)
+        return res;
+
+      if (!gtk_accessible_should_present (candidate))
         continue;
 
-      idx++;
+      res++;
+    }
+
+  return -1;
+}
+
+static int
+get_index_in_parent (GtkAccessible *accessible)
+{
+  GtkAccessible *parent = gtk_accessible_get_accessible_parent (accessible);
+
+  if (parent != NULL)
+    {
+      int res = get_index_in (parent, accessible);
+
+      g_object_unref (parent);
+
+      return res;
     }
 
   return -1;
@@ -385,58 +450,27 @@ static GVariant *
 get_parent_context_ref (GtkAccessible *accessible)
 {
   GVariant *res = NULL;
+  GtkAccessible *parent = gtk_accessible_get_accessible_parent (accessible);
 
-  if (GTK_IS_WIDGET (accessible))
+  if (parent == NULL)
     {
-      GtkWidget *widget = GTK_WIDGET (accessible);
-      GtkWidget *parent = gtk_widget_get_parent (widget);
+      GtkATContext *context = gtk_accessible_get_at_context (accessible);
+      GtkAtSpiContext *self = GTK_AT_SPI_CONTEXT (context);
 
-      if (parent == NULL)
-        {
-          GtkATContext *context = gtk_accessible_get_at_context (accessible);
-          GtkAtSpiContext *self = GTK_AT_SPI_CONTEXT (context);
+      res = gtk_at_spi_root_to_ref (self->root);
 
-          res = gtk_at_spi_root_to_ref (self->root);
-        }
-      else if (GTK_IS_STACK (parent))
-        {
-          GtkStackPage *page =
-            gtk_stack_get_page (GTK_STACK (parent), widget);
-          GtkATContext *parent_context =
-            gtk_accessible_get_at_context (GTK_ACCESSIBLE (page));
-
-          if (parent_context != NULL)
-            {
-              gtk_at_context_realize (parent_context);
-
-              res = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (parent_context));
-            }
-        }
-      else
-        {
-          GtkATContext *parent_context =
-            gtk_accessible_get_at_context (GTK_ACCESSIBLE (parent));
-
-          if (parent_context != NULL)
-            {
-              /* XXX: This realize() is needed otherwise opening a GtkPopover will
-               * emit a warning when getting the context's reference
-               */
-              gtk_at_context_realize (parent_context);
-
-              res = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (parent_context));
-            }
-        }
+      g_object_unref (context);
     }
-  else if (GTK_IS_STACK_PAGE (accessible))
+  else
     {
-      GtkWidget *parent =
-        gtk_widget_get_parent (gtk_stack_page_get_child (GTK_STACK_PAGE (accessible)));
-      GtkATContext *parent_context =
-        gtk_accessible_get_at_context (GTK_ACCESSIBLE (parent));
+      GtkATContext *parent_context = gtk_accessible_get_at_context (parent);
 
-      if (parent_context != NULL)
-        res = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (parent_context));
+      gtk_at_context_realize (parent_context);
+
+      res = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (parent_context));
+
+      g_object_unref (parent_context);
+      g_object_unref (parent);
     }
 
   if (res == NULL)
@@ -457,7 +491,7 @@ handle_accessible_method (GDBusConnection       *connection,
 {
   GtkAtSpiContext *self = user_data;
 
-  GTK_NOTE (A11Y, g_message ("handling %s on %s", method_name, object_path));
+  GTK_DEBUG (A11Y, "handling %s on %s", method_name, object_path);
 
   if (g_strcmp0 (method_name, "GetRole") == 0)
     {
@@ -517,50 +551,32 @@ handle_accessible_method (GDBusConnection       *connection,
     {
       GtkATContext *context = NULL;
       GtkAccessible *accessible;
-      int idx, real_idx = 0;
+      GtkAccessible *child = NULL;
+      int idx, presentable_idx;
 
       g_variant_get (parameters, "(i)", &idx);
 
       accessible = gtk_at_context_get_accessible (GTK_AT_CONTEXT (self));
 
-      if (GTK_IS_STACK_PAGE (accessible))
+      presentable_idx = 0;
+
+      for (child = gtk_accessible_get_first_accessible_child (accessible);
+           child != NULL;
+           child = gtk_accessible_get_next_accessible_sibling (child))
         {
-          if (idx == 0)
-            {
-              GtkWidget *child;
+          g_object_unref (child);
 
-              child = gtk_stack_page_get_child (GTK_STACK_PAGE (accessible));
-              if (gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
-                context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (child));
-            }
+          if (!gtk_accessible_should_present (child))
+            continue;
+
+          if (presentable_idx == idx)
+            break;
+
+          presentable_idx += 1;
         }
-      else if (GTK_IS_WIDGET (accessible))
-        {
-          GtkWidget *widget = GTK_WIDGET (accessible);
-          GtkWidget *child;
 
-          real_idx = 0;
-          for (child = gtk_widget_get_first_child (widget);
-               child;
-               child = gtk_widget_get_next_sibling (child))
-            {
-              if (!gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
-                continue;
-
-              if (real_idx == idx)
-                break;
-
-              real_idx += 1;
-            }
-
-          if (child)
-            {
-              if (GTK_IS_STACK (accessible))
-                context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (gtk_stack_get_page (GTK_STACK (accessible), child)));
-              else
-                context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (child));
-            }
-        }
+      if (child != NULL)
+        context = gtk_accessible_get_at_context (child);
 
       if (context == NULL)
         {
@@ -575,53 +591,37 @@ handle_accessible_method (GDBusConnection       *connection,
       gtk_at_context_realize (context);
 
       GVariant *ref = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (context));
-
       g_dbus_method_invocation_return_value (invocation, g_variant_new ("(@(so))", ref));
+
+      g_object_unref (context);
     }
   else if (g_strcmp0 (method_name, "GetChildren") == 0)
     {
       GVariantBuilder builder = G_VARIANT_BUILDER_INIT (G_VARIANT_TYPE ("a(so)"));
 
       GtkAccessible *accessible = gtk_at_context_get_accessible (GTK_AT_CONTEXT (self));
-      if (GTK_IS_WIDGET (accessible))
+      GtkAccessible *child = NULL;
+
+      for (child = gtk_accessible_get_first_accessible_child (accessible);
+           child != NULL;
+           child = gtk_accessible_get_next_accessible_sibling (child))
         {
-          GtkWidget *widget = GTK_WIDGET (accessible);
-          GtkWidget *child;
+          g_object_unref (child);
 
-          for (child = gtk_widget_get_first_child (widget);
-               child;
-               child = gtk_widget_get_next_sibling (child))
-            {
-              if (!gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
-                continue;
+          if (!gtk_accessible_should_present (child))
+            continue;
 
-              GtkATContext *context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (child));
+          GtkATContext *context = gtk_accessible_get_at_context (child);
 
-              /* Realize the child ATContext in order to get its ref */
-              gtk_at_context_realize (context);
+          /* Realize the child ATContext in order to get its ref */
+          gtk_at_context_realize (context);
 
-              GVariant *ref = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (context));
+          GVariant *ref = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (context));
 
-              if (ref != NULL)
-                g_variant_builder_add (&builder, "@(so)", ref);
-            }
-        }
-      else if (GTK_IS_STACK_PAGE (accessible))
-        {
-          GtkWidget *child = gtk_stack_page_get_child (GTK_STACK_PAGE (accessible));
+          if (ref != NULL)
+            g_variant_builder_add (&builder, "@(so)", ref);
 
-          if (gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
-            {
-              GtkATContext *context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (child));
-
-              /* Realize the child ATContext in order to get its ref */
-              gtk_at_context_realize (context);
-
-              GVariant *ref = gtk_at_spi_context_to_ref (GTK_AT_SPI_CONTEXT (context));
-
-              if (ref != NULL)
-                g_variant_builder_add (&builder, "@(so)", ref);
-            }
+          g_object_unref (context);
         }
 
       g_dbus_method_invocation_return_value (invocation, g_variant_new ("(a(so))", &builder));
@@ -662,7 +662,7 @@ handle_accessible_get_property (GDBusConnection       *connection,
 
   GtkAccessible *accessible = gtk_at_context_get_accessible (GTK_AT_CONTEXT (self));
 
-  GTK_NOTE (A11Y, g_message ("handling GetProperty %s on %s", property_name, object_path));
+  GTK_DEBUG (A11Y, "handling GetProperty %s on %s", property_name, object_path);
 
   if (g_strcmp0 (property_name, "Name") == 0)
     {
@@ -801,8 +801,13 @@ emit_property_changed (GtkAtSpiContext *self,
                        const char      *name,
                        GVariant        *value)
 {
+  GVariant *value_owned = g_variant_ref_sink (value);
+
   if (self->connection == NULL)
-    return;
+    {
+      g_variant_unref (value_owned);
+      return;
+    }
 
   g_dbus_connection_emit_signal (self->connection,
                                  NULL,
@@ -810,8 +815,9 @@ emit_property_changed (GtkAtSpiContext *self,
                                  "org.a11y.atspi.Event.Object",
                                  "PropertyChange",
                                  g_variant_new ("(siiva{sv})",
-                                                name, 0, 0, value, NULL),
+                                                name, 0, 0, value_owned, NULL),
                                  NULL);
+  g_variant_unref (value_owned);
 }
 
 static void
@@ -910,8 +916,6 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
 
   if (changed_states & GTK_ACCESSIBLE_STATE_CHANGE_HIDDEN)
     {
-      GtkWidget *parent;
-      GtkATContext *context;
       GtkAccessibleChildChange change;
 
       value = gtk_accessible_attribute_set_get_value (states, GTK_ACCESSIBLE_STATE_HIDDEN);
@@ -924,18 +928,19 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
         {
           gtk_at_spi_root_child_changed (self->root, change, accessible);
           emit_state_changed (self, "showing", gtk_boolean_accessible_value_get (value));
+          emit_state_changed (self, "visible", gtk_boolean_accessible_value_get (value));
         }
       else
         {
-          if (GTK_IS_WIDGET (accessible))
-            parent = gtk_widget_get_parent (GTK_WIDGET (accessible));
-          else if (GTK_IS_STACK_PAGE (accessible))
-            parent = gtk_widget_get_parent (gtk_stack_page_get_child (GTK_STACK_PAGE (accessible)));
-          else
-            g_assert_not_reached ();
+          GtkAccessible *parent =
+            gtk_accessible_get_accessible_parent (accessible);
+          GtkATContext *context =
+            gtk_accessible_get_at_context (parent);
 
-          context = gtk_accessible_get_at_context (GTK_ACCESSIBLE (parent));
           gtk_at_context_child_changed (context, change, accessible);
+
+          g_object_unref (context);
+          g_object_unref (parent);
         }
     }
 
@@ -949,22 +954,30 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
     {
       value = gtk_accessible_attribute_set_get_value (states, GTK_ACCESSIBLE_STATE_CHECKED);
 
-      switch (gtk_tristate_accessible_value_get (value))
+      if (value->value_class->type == GTK_ACCESSIBLE_VALUE_TYPE_TRISTATE)
+        {
+          switch (gtk_tristate_accessible_value_get (value))
+            {
+            case GTK_ACCESSIBLE_TRISTATE_TRUE:
+              emit_state_changed (self, "checked", TRUE);
+              emit_state_changed (self, "indeterminate", FALSE);
+              break;
+            case GTK_ACCESSIBLE_TRISTATE_MIXED:
+              emit_state_changed (self, "checked", FALSE);
+              emit_state_changed (self, "indeterminate", TRUE);
+              break;
+            case GTK_ACCESSIBLE_TRISTATE_FALSE:
+              emit_state_changed (self, "checked", FALSE);
+              emit_state_changed (self, "indeterminate", FALSE);
+              break;
+            default:
+              break;
+            }
+        }
+      else
         {
-        case GTK_ACCESSIBLE_TRISTATE_TRUE:
-          emit_state_changed (self, "checked", TRUE);
-          emit_state_changed (self, "indeterminate", FALSE);
-          break;
-        case GTK_ACCESSIBLE_TRISTATE_MIXED:
           emit_state_changed (self, "checked", FALSE);
           emit_state_changed (self, "indeterminate", TRUE);
-          break;
-        case GTK_ACCESSIBLE_TRISTATE_FALSE:
-          emit_state_changed (self, "checked", FALSE);
-          emit_state_changed (self, "indeterminate", FALSE);
-          break;
-        default:
-          break;
         }
     }
 
@@ -1007,22 +1020,31 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
   if (changed_states & GTK_ACCESSIBLE_STATE_CHANGE_PRESSED)
     {
       value = gtk_accessible_attribute_set_get_value (states, GTK_ACCESSIBLE_STATE_PRESSED);
-      switch (gtk_tristate_accessible_value_get (value))
+
+      if (value->value_class->type == GTK_ACCESSIBLE_VALUE_TYPE_TRISTATE)
+        {
+          switch (gtk_tristate_accessible_value_get (value))
+            {
+            case GTK_ACCESSIBLE_TRISTATE_TRUE:
+              emit_state_changed (self, "pressed", TRUE);
+              emit_state_changed (self, "indeterminate", FALSE);
+              break;
+            case GTK_ACCESSIBLE_TRISTATE_MIXED:
+              emit_state_changed (self, "pressed", FALSE);
+              emit_state_changed (self, "indeterminate", TRUE);
+              break;
+            case GTK_ACCESSIBLE_TRISTATE_FALSE:
+              emit_state_changed (self, "pressed", FALSE);
+              emit_state_changed (self, "indeterminate", FALSE);
+              break;
+            default:
+              break;
+            }
+        }
+      else
         {
-        case GTK_ACCESSIBLE_TRISTATE_TRUE:
-          emit_state_changed (self, "pressed", TRUE);
-          emit_state_changed (self, "indeterminate", FALSE);
-          break;
-        case GTK_ACCESSIBLE_TRISTATE_MIXED:
           emit_state_changed (self, "pressed", FALSE);
           emit_state_changed (self, "indeterminate", TRUE);
-          break;
-        case GTK_ACCESSIBLE_TRISTATE_FALSE:
-          emit_state_changed (self, "pressed", FALSE);
-          emit_state_changed (self, "indeterminate", FALSE);
-          break;
-        default:
-          break;
         }
     }
 
@@ -1038,6 +1060,15 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
         emit_state_changed (self, "selectable", FALSE);
     }
 
+  if (changed_states & GTK_ACCESSIBLE_STATE_CHANGE_VISITED)
+    {
+      value = gtk_accessible_attribute_set_get_value (states, GTK_ACCESSIBLE_STATE_VISITED);
+      if (value->value_class->type == GTK_ACCESSIBLE_VALUE_TYPE_BOOLEAN)
+        {
+          emit_state_changed (self, "visited",gtk_boolean_accessible_value_get (value));
+        }
+    }
+
   if (changed_properties & GTK_ACCESSIBLE_PROPERTY_CHANGE_READ_ONLY)
     {
       gboolean readonly;
@@ -1085,11 +1116,19 @@ gtk_at_spi_context_state_change (GtkATContext                *ctx,
     }
 
   if (changed_properties & GTK_ACCESSIBLE_PROPERTY_CHANGE_DESCRIPTION)
-    {
-      char *label = gtk_at_context_get_description (GTK_AT_CONTEXT (self));
-      GVariant *v = g_variant_new_take_string (label);
+  {
+      char *label = gtk_at_context_get_description (GTK_AT_CONTEXT (self));                                            
+      GVariant *v = g_variant_new_take_string (label);                                                                 
       emit_property_changed (self, "accessible-description", v);
     }
+
+  if (changed_properties & GTK_ACCESSIBLE_PROPERTY_CHANGE_VALUE_NOW)
+    {
+      value = gtk_accessible_attribute_set_get_value (properties, GTK_ACCESSIBLE_PROPERTY_VALUE_NOW);
+      emit_property_changed (self,
+                             "accessible-value",
+                             g_variant_new_double (gtk_number_accessible_value_get (value)));
+    }
 }
 
 static void
@@ -1131,7 +1170,7 @@ gtk_at_spi_context_platform_change (GtkATContext                *ctx,
       /* Orca tracks the window:activate and window:deactivate events on top
        * levels to decide whether to track other AT-SPI events
        */
-      if (gtk_accessible_get_accessible_role (accessible) == GTK_ACCESSIBLE_ROLE_WINDOW)
+      if (gtk_accessible_get_accessible_role (accessible) == GTK_ACCESSIBLE_ROLE_APPLICATION)
         {
           if (state)
             emit_window_event (self, "activate");
@@ -1146,29 +1185,9 @@ gtk_at_spi_context_bounds_change (GtkATContext *ctx)
 {
   GtkAtSpiContext *self = GTK_AT_SPI_CONTEXT (ctx);
   GtkAccessible *accessible = gtk_at_context_get_accessible (ctx);
-  GtkWidget *widget;
-  GtkWidget *parent;
-  double x, y;
-  int width, height;
-
-  if (!GTK_IS_WIDGET (accessible))
-    return;
-
-  widget = GTK_WIDGET (accessible);
-  if (!gtk_widget_get_realized (widget))
-    return;
-
-  parent = gtk_widget_get_parent (widget);
-
-  if (parent)
-    gtk_widget_translate_coordinates (widget, parent, 0., 0., &x, &y);
-  else
-    x = y = 0.;
-
-  width = gtk_widget_get_width (widget);
-  height = gtk_widget_get_height (widget);
-
-  emit_bounds_changed (self, (int)x, (int)y, width, height);
+  int x, y, width, height;
+  if (gtk_accessible_get_bounds (accessible, &x, &y, &width, &height))
+    emit_bounds_changed (self, x, y, width, height);
 }
 
 static void
@@ -1179,98 +1198,25 @@ gtk_at_spi_context_child_change (GtkATContext             *ctx,
   GtkAtSpiContext *self = GTK_AT_SPI_CONTEXT (ctx);
   GtkAccessible *accessible = gtk_at_context_get_accessible (ctx);
   GtkATContext *child_context = gtk_accessible_get_at_context (child);
-  GtkWidget *parent_widget;
-  GtkWidget *child_widget;
-  int idx = 0;
-
-  if (!GTK_IS_WIDGET (accessible))
-    return;
 
   if (child_context == NULL)
     return;
 
-  /* handle the stack page special case */
-  if (GTK_IS_WIDGET (child) &&
-      GTK_IS_STACK (gtk_widget_get_parent (GTK_WIDGET (child))))
+  GtkAccessible *parent = gtk_accessible_get_accessible_parent (child);
+  int idx = 0;
+
+  if (parent == NULL)
     {
-      GtkWidget *stack;
-      GtkStackPage *page;
-      GListModel *pages;
-
-      stack = gtk_widget_get_parent (GTK_WIDGET (child));
-      page = gtk_stack_get_page (GTK_STACK (stack), GTK_WIDGET (child));
-      pages = G_LIST_MODEL (gtk_stack_get_pages (GTK_STACK (stack)));
-      idx = 0;
-      for (guint i = 0; i < g_list_model_get_n_items (pages); i++)
-        {
-          GtkStackPage *item = g_list_model_get_item (pages, i);
-
-          g_object_unref (item);
-
-          if (!gtk_accessible_should_present (GTK_ACCESSIBLE (item)))
-            continue;
-
-          if (item == page)
-            break;
-
-          idx++;
-        }
-      g_object_unref (pages);
-
-      if (change & GTK_ACCESSIBLE_CHILD_CHANGE_ADDED)
-        {
-          emit_children_changed (GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (stack))),
-                                 GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (page))),
-                                 idx,
-                                 GTK_ACCESSIBLE_CHILD_STATE_ADDED);
-
-          emit_children_changed (GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (page))),
-                                 GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (child)),
-                                 0,
-                                 GTK_ACCESSIBLE_CHILD_STATE_ADDED);
-        }
-
-      if (change & GTK_ACCESSIBLE_CHILD_CHANGE_REMOVED)
-        {
-          emit_children_changed (GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (page))),
-                                 GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (child)),
-                                 0,
-                                 GTK_ACCESSIBLE_CHILD_STATE_REMOVED);
-          emit_children_changed (GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (stack))),
-                                 GTK_AT_SPI_CONTEXT (gtk_accessible_get_at_context (GTK_ACCESSIBLE (page))),
-                                 idx,
-                                 GTK_ACCESSIBLE_CHILD_STATE_REMOVED);
-
-        }
-
-      return;
+      idx = -1;
     }
-
-  parent_widget = GTK_WIDGET (accessible);
-
-  if (GTK_IS_STACK_PAGE (child))
-    child_widget = gtk_stack_page_get_child (GTK_STACK_PAGE (child));
-  else
-    child_widget = GTK_WIDGET (child);
-
-  if (gtk_widget_get_parent (child_widget) != parent_widget)
+  else if (parent == accessible)
     {
-      idx = 0;
+      idx = get_index_in (accessible, child);
+      g_object_unref (parent);
     }
   else
     {
-      for (GtkWidget *iter = gtk_widget_get_first_child (parent_widget);
-           iter != NULL;
-           iter = gtk_widget_get_next_sibling (iter))
-        {
-          if (!gtk_accessible_should_present (GTK_ACCESSIBLE (iter)))
-            continue;
-
-          if (iter == child_widget)
-            break;
-
-          idx += 1;
-        }
+      g_object_unref (parent);
     }
 
   if (change & GTK_ACCESSIBLE_CHILD_CHANGE_ADDED)
@@ -1283,6 +1229,8 @@ gtk_at_spi_context_child_change (GtkATContext             *ctx,
                            GTK_AT_SPI_CONTEXT (child_context),
                            idx,
                            GTK_ACCESSIBLE_CHILD_STATE_REMOVED);
+
+  g_object_unref (child_context);
 }
 /* }}} */
 /* {{{ D-Bus Registration */
@@ -1399,9 +1347,9 @@ gtk_at_spi_context_register_object (GtkAtSpiContext *self)
 
   self->interfaces = g_variant_ref_sink (g_variant_builder_end (&interfaces));
 
-  GTK_NOTE (A11Y, g_message ("Registered %d interfaces on object path '%s'",
-                             self->n_registered_objects,
-                             self->context_path));
+  GTK_DEBUG (A11Y, "Registered %d interfaces on object path '%s'",
+                   self->n_registered_objects,
+                   self->context_path);
 }
 
 static void
@@ -1418,7 +1366,7 @@ gtk_at_spi_context_unregister_object (GtkAtSpiContext *self)
   g_clear_pointer (&self->interfaces, g_variant_unref);
 }
 /* }}} */
-/* {{{ GObject boilerplate */ 
+/* {{{ GObject boilerplate */
 static void
 gtk_at_spi_context_finalize (GObject *gobject)
 {
@@ -1443,6 +1391,22 @@ gtk_at_spi_context_constructed (GObject *gobject)
 
 static const char *get_bus_address (GdkDisplay *display);
 
+static void
+register_object (GtkAtSpiRoot *root,
+                 GtkAtSpiContext *context)
+{
+  GtkAccessible *accessible = gtk_at_context_get_accessible (GTK_AT_CONTEXT (context));
+
+  gtk_atspi_connect_text_signals (accessible,
+                                  (GtkAtspiTextChangedCallback *)emit_text_changed,
+                                  (GtkAtspiTextSelectionCallback *)emit_text_selection_changed,
+                                  context);
+  gtk_atspi_connect_selection_signals (accessible,
+                                       (GtkAtspiSelectionCallback *)emit_selection_changed,
+                                       context);
+  gtk_at_spi_context_register_object (context);
+}
+
 static void
 gtk_at_spi_context_realize (GtkATContext *context)
 {
@@ -1489,11 +1453,10 @@ gtk_at_spi_context_realize (GtkATContext *context)
   if (self->connection == NULL)
     return;
 
-  GtkAccessible *accessible = gtk_at_context_get_accessible (context);
-
 #ifdef G_ENABLE_DEBUG
   if (GTK_DEBUG_CHECK (A11Y))
     {
+      GtkAccessible *accessible = gtk_at_context_get_accessible (context);
       GtkAccessibleRole role = gtk_at_context_get_accessible_role (context);
       char *role_name = g_enum_to_string (GTK_TYPE_ACCESSIBLE_ROLE, role);
       g_message ("Realizing ATSPI context “%s” for accessible “%s”, with role: “%s”",
@@ -1504,16 +1467,7 @@ gtk_at_spi_context_realize (GtkATContext *context)
     }
 #endif
 
-  gtk_atspi_connect_text_signals (accessible,
-                                  (GtkAtspiTextChangedCallback *)emit_text_changed,
-                                  (GtkAtspiTextSelectionCallback *)emit_text_selection_changed,
-                                  self);
-  gtk_atspi_connect_selection_signals (accessible,
-                                       (GtkAtspiSelectionCallback *)emit_selection_changed,
-                                       self);
-  gtk_at_spi_context_register_object (self);
-
-  gtk_at_spi_root_queue_register (self->root, self);
+  gtk_at_spi_root_queue_register (self->root, self, register_object);
 }
 
 static void
@@ -1522,9 +1476,9 @@ gtk_at_spi_context_unrealize (GtkATContext *context)
   GtkAtSpiContext *self = GTK_AT_SPI_CONTEXT (context);
   GtkAccessible *accessible = gtk_at_context_get_accessible (context);
 
-  GTK_NOTE (A11Y, g_message ("Unrealizing ATSPI context at '%s' for accessible '%s'",
-                             self->context_path,
-                             G_OBJECT_TYPE_NAME (accessible)));
+  GTK_DEBUG (A11Y, "Unrealizing ATSPI context at '%s' for accessible '%s'",
+                   self->context_path,
+                   G_OBJECT_TYPE_NAME (accessible));
 
   /* Notify ATs that the accessible object is going away */
   emit_defunct (self);
@@ -1565,7 +1519,7 @@ gtk_at_spi_context_init (GtkAtSpiContext *self)
 static char *
 get_bus_address_x11 (GdkDisplay *display)
 {
-  GTK_NOTE (A11Y, g_message ("Acquiring a11y bus via X11..."));
+  GTK_DEBUG (A11Y, "Acquiring a11y bus via X11...");
 
   Display *xdisplay = gdk_x11_display_get_xdisplay (display);
   Atom type_return;
@@ -1596,14 +1550,14 @@ get_bus_address_x11 (GdkDisplay *display)
 static char *
 get_bus_address_dbus (GdkDisplay *display)
 {
-  GTK_NOTE (A11Y, g_message ("Acquiring a11y bus via DBus..."));
+  GTK_DEBUG (A11Y, "Acquiring a11y bus via DBus...");
 
   GError *error = NULL;
   GDBusConnection *connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
 
   if (error != NULL)
     {
-      GTK_NOTE (A11Y, g_message ("Unable to acquire session bus: %s", error->message));
+      g_warning ("Unable to acquire session bus: %s", error->message);
       g_error_free (error);
       return NULL;
     }
@@ -1620,8 +1574,10 @@ get_bus_address_dbus (GdkDisplay *display)
                                   &error);
   if (error != NULL)
     {
-      GTK_NOTE (A11Y, g_message ("Unable to acquire the address of the accessibility bus: %s",
-                                 error->message));
+      g_warning ("Unable to acquire the address of the accessibility bus: %s. "
+                 "If you are attempting to run GTK without a11y support, "
+                 "GTK_A11Y should be set to 'none'.",
+                 error->message);
       g_error_free (error);
     }
 
@@ -1654,7 +1610,7 @@ get_bus_address (GdkDisplay *display)
   bus_address = g_getenv ("AT_SPI_BUS_ADDRESS");
   if (bus_address != NULL && *bus_address != '\0')
     {
-      GTK_NOTE (A11Y, g_message ("Using ATSPI bus address from environment: %s", bus_address));
+      GTK_DEBUG (A11Y, "Using ATSPI bus address from environment: %s", bus_address);
       g_object_set_data_full (G_OBJECT (display), "-gtk-atspi-bus-address",
                               g_strdup (bus_address),
                               g_free);
@@ -1668,7 +1624,7 @@ get_bus_address (GdkDisplay *display)
         {
           char *addr = get_bus_address_dbus (display);
 
-          GTK_NOTE (A11Y, g_message ("Using ATSPI bus address from D-Bus: %s", addr));
+          GTK_DEBUG (A11Y, "Using ATSPI bus address from D-Bus: %s", addr);
           g_object_set_data_full (G_OBJECT (display), "-gtk-atspi-bus-address",
                                   addr,
                                   g_free);
@@ -1687,11 +1643,11 @@ get_bus_address (GdkDisplay *display)
           if (addr == NULL)
             {
               addr = get_bus_address_x11 (display);
-              GTK_NOTE (A11Y, g_message ("Using ATSPI bus address from X11: %s", addr));
+              GTK_DEBUG (A11Y, "Using ATSPI bus address from X11: %s", addr);
             }
           else
             {
-              GTK_NOTE (A11Y, g_message ("Using ATSPI bus address from D-Bus: %s", addr));
+              GTK_DEBUG (A11Y, "Using ATSPI bus address from D-Bus: %s", addr);
             }
 
           g_object_set_data_full (G_OBJECT (display), "-gtk-atspi-bus-address",
@@ -1704,6 +1660,12 @@ get_bus_address (GdkDisplay *display)
 #endif
 
 out:
+
+  if (bus_address == NULL)
+    g_object_set_data_full (G_OBJECT (display), "-gtk-atspi-bus-address",
+                            g_strdup (""),
+                            g_free);
+
   return bus_address;
 }
 
@@ -1822,12 +1784,8 @@ gtk_at_spi_context_get_index_in_parent (GtkAtSpiContext *self)
 
   if (GTK_IS_ROOT (accessible))
     idx = get_index_in_toplevels (GTK_WIDGET (accessible));
-  else if (GTK_IS_STACK_PAGE (accessible))
-    idx = get_index_in_parent (gtk_stack_page_get_child (GTK_STACK_PAGE (accessible)));
-  else if (GTK_IS_STACK (gtk_widget_get_parent (GTK_WIDGET (accessible))))
-    idx = 1;
   else
-    idx = get_index_in_parent (GTK_WIDGET (accessible));
+    idx = get_index_in_parent  (accessible);
 
   return idx;
 }
@@ -1838,26 +1796,19 @@ gtk_at_spi_context_get_child_count (GtkAtSpiContext *self)
   g_return_val_if_fail (GTK_IS_AT_SPI_CONTEXT (self), -1);
 
   GtkAccessible *accessible = gtk_at_context_get_accessible (GTK_AT_CONTEXT (self));
-  int n_children = -1;
+  int n_children = 0;
 
-  if (GTK_IS_WIDGET (accessible))
+  GtkAccessible *child = NULL;
+  for (child = gtk_accessible_get_first_accessible_child (accessible);
+       child != NULL;
+       child = gtk_accessible_get_next_accessible_sibling (child))
     {
-      GtkWidget *child;
+      g_object_unref (child);
 
-      n_children = 0;
-      for (child = gtk_widget_get_first_child (GTK_WIDGET (accessible));
-           child;
-           child = gtk_widget_get_next_sibling (child))
-        {
-          if (!gtk_accessible_should_present (GTK_ACCESSIBLE (child)))
-            continue;
+      if (!gtk_accessible_should_present (child))
+        continue;
 
-          n_children++;
-        }
-    }
-  else if (GTK_IS_STACK_PAGE (accessible))
-    {
-      n_children = 1;
+      n_children += 1;
     }
 
   return n_children;
