@@ -22,7 +22,6 @@
 #include "gtkdropdown.h"
 
 #include "gtkbuiltiniconprivate.h"
-#include "gtkintl.h"
 #include "gtklistview.h"
 #include "gtklistitemfactory.h"
 #include "gtksignallistitemfactory.h"
@@ -45,6 +44,7 @@
 #include "gtkbuilderprivate.h"
 #include "gtkstringlist.h"
 #include "gtkbox.h"
+#include "gtktypebuiltins.h"
 
 /**
  * GtkDropDown:
@@ -54,11 +54,17 @@
  *
  * ![An example GtkDropDown](drop-down.png)
  *
- * The `GtkDropDown` displays the selected choice.
+ * The `GtkDropDown` displays the [selected][property@Gtk.DropDown:selected]
+ * choice.
  *
  * The options are given to `GtkDropDown` in the form of `GListModel`
  * and how the individual options are represented is determined by
- * a [class@Gtk.ListItemFactory]. The default factory displays simple strings.
+ * a [class@Gtk.ListItemFactory]. The default factory displays simple strings,
+ * and adds a checkmark to the selected item in the popup.
+ *
+ * To set your own factory, use [method@Gtk.DropDown.set_factory]. It is
+ * possible to use a separate factory for the items in the popup, with
+ * [method@Gtk.DropDown.set_list_factory].
  *
  * `GtkDropDown` knows how to obtain strings from the items in a
  * [class@Gtk.StringList]; for other models, you have to provide an expression
@@ -68,12 +74,31 @@
  * useful if the list of options is long. To enable the search entry,
  * use [method@Gtk.DropDown.set_enable_search].
  *
- * # CSS nodes
+ * Here is a UI definition example for `GtkDropDown` with a simple model:
+ *
+ * ```xml
+ * <object class="GtkDropDown">
+ *   <property name="model">
+ *     <object class="GtkStringList">
+ *       <items>
+ *         <item translatable="yes">Factory</item>
+ *         <item translatable="yes">Home</item>
+ *         <item translatable="yes">Subway</item>
+ *       </items>
+ *     </object>
+ *   </property>
+ * </object>
+ * ```
+ *
+ * To learn more about the list widget framework, see the
+ * [overview](section-list-widget.html).
+ *
+ * ## CSS nodes
  *
  * `GtkDropDown` has a single CSS node with name dropdown,
  * with the button and popover nodes as children.
  *
- * # Accessibility
+ * ## Accessibility
  *
  * `GtkDropDown` uses the %GTK_ACCESSIBLE_ROLE_COMBO_BOX role.
  */
@@ -82,8 +107,11 @@ struct _GtkDropDown
 {
   GtkWidget parent_instance;
 
+  gboolean uses_default_factory;
+  gboolean uses_default_list_factory;
   GtkListItemFactory *factory;
   GtkListItemFactory *list_factory;
+  GtkListItemFactory *header_factory;
   GListModel *model;
   GtkSelectionModel *selection;
   GListModel *filter_model;
@@ -102,6 +130,8 @@ struct _GtkDropDown
 
   GtkExpression *expression;
 
+  GtkStringFilterMatchMode search_match_mode;
+
   guint enable_search : 1;
   guint show_arrow : 1;
 };
@@ -115,6 +145,7 @@ enum
 {
   PROP_0,
   PROP_FACTORY,
+  PROP_HEADER_FACTORY,
   PROP_LIST_FACTORY,
   PROP_MODEL,
   PROP_SELECTED,
@@ -122,6 +153,7 @@ enum
   PROP_ENABLE_SEARCH,
   PROP_EXPRESSION,
   PROP_SHOW_ARROW,
+  PROP_SEARCH_MATCH_MODE,
 
   N_PROPS
 };
@@ -147,6 +179,10 @@ button_toggled (GtkWidget *widget,
     gtk_popover_popup (GTK_POPOVER (self->popup));
   else
     gtk_popover_popdown (GTK_POPOVER (self->popup));
+
+  gtk_accessible_update_state (GTK_ACCESSIBLE (self),
+                               GTK_ACCESSIBLE_STATE_EXPANDED, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)),
+                               -1);
 }
 
 static void
@@ -184,11 +220,9 @@ selection_changed (GtkSingleSelection *selection,
 {
   GtkDropDown *self = data;
   guint selected;
-  gpointer item;
   GtkFilter *filter;
 
   selected = gtk_single_selection_get_selected (GTK_SINGLE_SELECTION (self->selection));
-  item = gtk_single_selection_get_selected_item (GTK_SINGLE_SELECTION (self->selection));
 
   if (selected == GTK_INVALID_LIST_POSITION)
     {
@@ -197,7 +231,14 @@ selection_changed (GtkSingleSelection *selection,
   else
     {
       gtk_stack_set_visible_child_name (GTK_STACK (self->button_stack), "item");
-      gtk_list_item_widget_update (GTK_LIST_ITEM_WIDGET (self->button_item), selected, item, FALSE);
+    }
+
+  if (selected != gtk_list_item_base_get_position (GTK_LIST_ITEM_BASE (self->button_item)))
+    {
+      gtk_list_item_base_update (GTK_LIST_ITEM_BASE (self->button_item),
+                                 selected,
+                                 gtk_single_selection_get_selected_item (GTK_SINGLE_SELECTION (self->selection)),
+                                 FALSE);
     }
 
   /* reset the filter so positions are 1-1 */
@@ -207,6 +248,26 @@ selection_changed (GtkSingleSelection *selection,
   gtk_single_selection_set_selected (GTK_SINGLE_SELECTION (self->popup_selection), selected);
 
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SELECTED]);
+}
+
+static void
+selection_item_changed (GtkSingleSelection *selection,
+                        GParamSpec         *pspec,
+                        gpointer            data)
+{
+  GtkDropDown *self = data;
+  gpointer item;
+
+  item = gtk_single_selection_get_selected_item (GTK_SINGLE_SELECTION (self->selection));
+
+  if (item != gtk_list_item_base_get_item (GTK_LIST_ITEM_BASE (self->button_item)))
+    {
+      gtk_list_item_base_update (GTK_LIST_ITEM_BASE (self->button_item),
+                                 gtk_single_selection_get_selected (GTK_SINGLE_SELECTION (self->selection)),
+                                 item,
+                                 FALSE);
+    }
+
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SELECTED_ITEM]);
 }
 
@@ -226,7 +287,7 @@ update_filter (GtkDropDown *self)
       if (self->expression)
         {
           filter = GTK_FILTER (gtk_string_filter_new (gtk_expression_ref (self->expression)));
-          gtk_string_filter_set_match_mode (GTK_STRING_FILTER (filter), GTK_STRING_FILTER_MATCH_MODE_PREFIX);
+          gtk_string_filter_set_match_mode (GTK_STRING_FILTER (filter), self->search_match_mode);
         }
       else
         filter = GTK_FILTER (gtk_every_filter_new ());
@@ -275,13 +336,17 @@ gtk_drop_down_dispose (GObject *object)
 
   g_clear_object (&self->model);
   if (self->selection)
-    g_signal_handlers_disconnect_by_func (self->selection, selection_changed, self);
+    {
+      g_signal_handlers_disconnect_by_func (self->selection, selection_changed, self);
+      g_signal_handlers_disconnect_by_func (self->selection, selection_item_changed, self);
+    }
   g_clear_object (&self->filter_model);
   g_clear_pointer (&self->expression, gtk_expression_unref);
   g_clear_object (&self->selection);
   g_clear_object (&self->popup_selection);
   g_clear_object (&self->factory);
   g_clear_object (&self->list_factory);
+  g_clear_object (&self->header_factory);
 
   G_OBJECT_CLASS (gtk_drop_down_parent_class)->dispose (object);
 }
@@ -300,6 +365,10 @@ gtk_drop_down_get_property (GObject    *object,
       g_value_set_object (value, self->factory);
       break;
 
+    case PROP_HEADER_FACTORY:
+      g_value_set_object (value, self->header_factory);
+      break;
+
     case PROP_LIST_FACTORY:
       g_value_set_object (value, self->list_factory);
       break;
@@ -328,6 +397,10 @@ gtk_drop_down_get_property (GObject    *object,
       g_value_set_boolean (value, gtk_drop_down_get_show_arrow (self));
       break;
 
+    case PROP_SEARCH_MATCH_MODE:
+      g_value_set_enum (value, gtk_drop_down_get_search_match_mode (self));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -348,6 +421,10 @@ gtk_drop_down_set_property (GObject      *object,
       gtk_drop_down_set_factory (self, g_value_get_object (value));
       break;
 
+    case PROP_HEADER_FACTORY:
+      gtk_drop_down_set_header_factory (self, g_value_get_object (value));
+      break;
+
     case PROP_LIST_FACTORY:
       gtk_drop_down_set_list_factory (self, g_value_get_object (value));
       break;
@@ -372,6 +449,11 @@ gtk_drop_down_set_property (GObject      *object,
       gtk_drop_down_set_show_arrow (self, g_value_get_boolean (value));
       break;
 
+    case PROP_SEARCH_MATCH_MODE:
+      gtk_drop_down_set_search_match_mode (self, g_value_get_enum (value));
+      break;
+
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -432,6 +514,27 @@ gtk_drop_down_grab_focus (GtkWidget *widget)
   return gtk_widget_grab_focus (self->button);
 }
 
+static void
+gtk_drop_down_root (GtkWidget *widget)
+{
+  GtkDropDown *self = GTK_DROP_DOWN (widget);
+
+  GTK_WIDGET_CLASS (gtk_drop_down_parent_class)->root (widget);
+
+  if (self->factory)
+    gtk_list_factory_widget_set_factory (GTK_LIST_FACTORY_WIDGET (self->button_item), self->factory);
+}
+
+static void
+gtk_drop_down_unroot (GtkWidget *widget)
+{
+  GtkDropDown *self = GTK_DROP_DOWN (widget);
+
+  if (self->factory)
+    gtk_list_factory_widget_set_factory (GTK_LIST_FACTORY_WIDGET (self->button_item), NULL);
+
+  GTK_WIDGET_CLASS (gtk_drop_down_parent_class)->unroot (widget);
+}
 
 static void
 gtk_drop_down_class_init (GtkDropDownClass *klass)
@@ -447,6 +550,8 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
   widget_class->size_allocate = gtk_drop_down_size_allocate;
   widget_class->focus = gtk_drop_down_focus;
   widget_class->grab_focus = gtk_drop_down_grab_focus;
+  widget_class->root = gtk_drop_down_root;
+  widget_class->unroot = gtk_drop_down_unroot;
 
   /**
    * GtkDropDown:factory: (attributes org.gtk.Property.get=gtk_drop_down_get_factory org.gtk.Property.set=gtk_drop_down_set_factory)
@@ -454,9 +559,19 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * Factory for populating list items.
    */
   properties[PROP_FACTORY] =
-    g_param_spec_object ("factory",
-                         P_("Factory"),
-                         P_("Factory for populating list items"),
+    g_param_spec_object ("factory", NULL, NULL,
+                         GTK_TYPE_LIST_ITEM_FACTORY,
+                         G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GtkDropDown:header-factory: (attributes org.gtk.Property.get=gtk_drop_down_get_header_factory org.gtk.Property.set=gtk_drop_down_set_header_factory)
+   *
+   * The factory for creating header widgets for the popup.
+   *
+   * Since: 4.12
+   */
+  properties[PROP_HEADER_FACTORY] =
+    g_param_spec_object ("header-factory", NULL, NULL,
                          GTK_TYPE_LIST_ITEM_FACTORY,
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
@@ -468,9 +583,7 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * If this is not set, [property@Gtk.DropDown:factory] is used.
    */
   properties[PROP_LIST_FACTORY] =
-    g_param_spec_object ("list-factory",
-                         P_("List Factory"),
-                         P_("Factory for populating list items"),
+    g_param_spec_object ("list-factory", NULL, NULL,
                          GTK_TYPE_LIST_ITEM_FACTORY,
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
@@ -480,9 +593,7 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * Model for the displayed items.
    */
   properties[PROP_MODEL] =
-    g_param_spec_object ("model",
-                         P_("Model"),
-                         P_("Model for the displayed items"),
+    g_param_spec_object ("model", NULL, NULL,
                          G_TYPE_LIST_MODEL,
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
@@ -495,11 +606,9 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * %GTK_INVALID_LIST_POSITION.
    */
   properties[PROP_SELECTED] =
-    g_param_spec_uint ("selected",
-                         P_("Selected"),
-                         P_("Position of the selected item"),
-                         0, G_MAXUINT, GTK_INVALID_LIST_POSITION,
-                         G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+    g_param_spec_uint ("selected", NULL, NULL,
+                       0, G_MAXUINT, GTK_INVALID_LIST_POSITION,
+                       G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
    * GtkDropDown:selected-item: (attributes org.gtk.Property.get=gtk_drop_down_get_selected_item)
@@ -507,11 +616,9 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * The selected item.
    */
   properties[PROP_SELECTED_ITEM] =
-    g_param_spec_object ("selected-item",
-                       P_("Selected Item"),
-                       P_("The selected item"),
-                       G_TYPE_OBJECT,
-                       G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+    g_param_spec_object ("selected-item", NULL, NULL,
+                         G_TYPE_OBJECT,
+                         G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
   /**
    * GtkDropDown:enable-search: (attributes org.gtk.Property.get=gtk_drop_down_get_enable_search org.gtk.Property.set=gtk_drop_down_set_enable_search)
@@ -522,9 +629,7 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * to be set.
    */
   properties[PROP_ENABLE_SEARCH] =
-    g_param_spec_boolean  ("enable-search",
-                         P_("Enable search"),
-                         P_("Whether to show a search entry in the popup"),
+    g_param_spec_boolean  ("enable-search", NULL, NULL,
                          FALSE,
                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
@@ -539,9 +644,7 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * used to bind strings to labels produced by a default factory.
    */
   properties[PROP_EXPRESSION] =
-    gtk_param_spec_expression ("expression",
-                               P_("Expression"),
-                               P_("Expression to determine strings to search for"),
+    gtk_param_spec_expression ("expression", NULL, NULL,
                                G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
@@ -552,12 +655,23 @@ gtk_drop_down_class_init (GtkDropDownClass *klass)
    * Since: 4.6
    */
   properties[PROP_SHOW_ARROW] =
-    g_param_spec_boolean  ("show-arrow",
-                           P_("Show arrow"),
-                           P_("Whether to show an arrow within the widget"),
+    g_param_spec_boolean  ("show-arrow", NULL, NULL,
                            TRUE,
                            G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * GtkDropDown:search-match-mode: (attributes org.gtk.Property.get=gtk_drop_down_get_search_match_mode org.gtk.Property.set=gtk_drop_down_set_search_match_mode)
+   *
+   * The match mode for the search filter.
+   *
+   * Since: 4.12
+   */
+  properties[PROP_SEARCH_MATCH_MODE] =
+    g_param_spec_enum  ("search-match-mode", NULL, NULL,
+                           GTK_TYPE_STRING_FILTER_MATCH_MODE,
+                           GTK_STRING_FILTER_MATCH_MODE_PREFIX,
+                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
   g_object_class_install_properties (gobject_class, N_PROPS, properties);
 
   /**
@@ -615,7 +729,10 @@ setup_item (GtkSignalListItemFactory *factory,
   label = gtk_label_new (NULL);
   gtk_label_set_xalign (GTK_LABEL (label), 0.0);
   gtk_box_append (GTK_BOX (box), label);
-  icon = gtk_image_new_from_icon_name ("object-select-symbolic");
+  icon = g_object_new (GTK_TYPE_IMAGE,
+                       "icon-name", "object-select-symbolic",
+                       "accessible-role", GTK_ACCESSIBLE_ROLE_PRESENTATION,
+                       NULL);
   gtk_box_append (GTK_BOX (box), icon);
   gtk_list_item_set_child (list_item, box);
 }
@@ -674,14 +791,14 @@ bind_item (GtkSignalListItemFactory *factory,
 
   if (gtk_widget_get_ancestor (box, GTK_TYPE_POPOVER) == self->popup)
     {
-      gtk_widget_show (icon);
+      gtk_widget_set_visible (icon, TRUE);
       g_signal_connect (self, "notify::selected-item",
                         G_CALLBACK (selected_item_changed), list_item);
       selected_item_changed (self, NULL, list_item);
     }
   else
     {
-      gtk_widget_hide (icon);
+      gtk_widget_set_visible (icon, FALSE);
     }
 }
 
@@ -707,6 +824,10 @@ set_default_factory (GtkDropDown *self)
   g_signal_connect (factory, "unbind", G_CALLBACK (unbind_item), self);
 
   gtk_drop_down_set_factory (self, factory);
+  self->uses_default_factory = TRUE;
+
+  if (self->uses_default_list_factory)
+    gtk_drop_down_set_list_factory (self, NULL);
 
   g_object_unref (factory);
 }
@@ -722,6 +843,8 @@ gtk_drop_down_init (GtkDropDown *self)
   self->show_arrow = gtk_widget_get_visible (self->arrow);
 
   set_default_factory (self);
+
+  self->search_match_mode = GTK_STRING_FILTER_MATCH_MODE_PREFIX;
 }
 
 /**
@@ -807,7 +930,10 @@ gtk_drop_down_set_model (GtkDropDown *self,
       gtk_list_view_set_model (GTK_LIST_VIEW (self->popup_list), NULL);
 
       if (self->selection)
-        g_signal_handlers_disconnect_by_func (self->selection, selection_changed, self);
+        {
+          g_signal_handlers_disconnect_by_func (self->selection, selection_changed, self);
+          g_signal_handlers_disconnect_by_func (self->selection, selection_item_changed, self);
+        }
 
       g_clear_object (&self->selection);
       g_clear_object (&self->filter_model);
@@ -833,7 +959,9 @@ gtk_drop_down_set_model (GtkDropDown *self,
       g_object_unref (selection);
 
       g_signal_connect (self->selection, "notify::selected", G_CALLBACK (selection_changed), self);
+      g_signal_connect (self->selection, "notify::selected-item", G_CALLBACK (selection_item_changed), self);
       selection_changed (GTK_SINGLE_SELECTION (self->selection), NULL, self);
+      selection_item_changed (GTK_SINGLE_SELECTION (self->selection), NULL, self);
     }
 
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_MODEL]);
@@ -876,13 +1004,62 @@ gtk_drop_down_set_factory (GtkDropDown        *self,
   if (!g_set_object (&self->factory, factory))
     return;
 
-  gtk_list_item_widget_set_factory (GTK_LIST_ITEM_WIDGET (self->button_item), factory);
+  if (gtk_widget_get_root (GTK_WIDGET (self)))
+    gtk_list_factory_widget_set_factory (GTK_LIST_FACTORY_WIDGET (self->button_item), factory);
+
   if (self->list_factory == NULL)
-    gtk_list_view_set_factory (GTK_LIST_VIEW (self->popup_list), factory);
+    {
+      gtk_list_view_set_factory (GTK_LIST_VIEW (self->popup_list), factory);
+      self->uses_default_list_factory = TRUE;
+    }
+
+  self->uses_default_factory = factory != NULL;
 
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_FACTORY]);
 }
 
+/**
+ * gtk_drop_down_get_header_factory: (attributes org.gtk.Method.get_property=header-factory)
+ * @self: a `GtkDropDown`
+ *
+ * Gets the factory that's currently used to create header widgets for the popup.
+ *
+ * Returns: (nullable) (transfer none): The factory in use
+ *
+ * Since: 4.12
+ */
+GtkListItemFactory *
+gtk_drop_down_get_header_factory (GtkDropDown *self)
+{
+  g_return_val_if_fail (GTK_IS_DROP_DOWN (self), NULL);
+
+  return self->header_factory;
+}
+
+/**
+ * gtk_drop_down_set_header_factory: (attributes org.gtk.Method.set_property=header-factory)
+ * @self: a `GtkDropDown`
+ * @factory: (nullable) (transfer none): the factory to use
+ *
+ * Sets the `GtkListItemFactory` to use for creating header widgets for the popup.
+ *
+ * Since: 4.12
+ */
+void
+gtk_drop_down_set_header_factory (GtkDropDown        *self,
+                                  GtkListItemFactory *factory)
+{
+  g_return_if_fail (GTK_IS_DROP_DOWN (self));
+  g_return_if_fail (factory == NULL || GTK_LIST_ITEM_FACTORY (factory));
+
+  if (!g_set_object (&self->header_factory, factory))
+    return;
+
+  gtk_list_view_set_header_factory (GTK_LIST_VIEW (self->popup_list), self->header_factory);
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_HEADER_FACTORY]);
+}
+
 /**
  * gtk_drop_down_get_list_factory: (attributes org.gtk.Method.get_property=list-factory)
  * @self: a `GtkDropDown`
@@ -921,6 +1098,8 @@ gtk_drop_down_set_list_factory (GtkDropDown        *self,
   else
     gtk_list_view_set_factory (GTK_LIST_VIEW (self->popup_list), self->factory);
 
+  self->uses_default_list_factory = factory != NULL;
+
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_LIST_FACTORY]);
 }
 
@@ -1058,6 +1237,9 @@ gtk_drop_down_set_expression (GtkDropDown   *self,
   if (self->expression)
     gtk_expression_ref (self->expression);
 
+  if (self->uses_default_factory)
+    set_default_factory (self);
+
   update_filter (self);
 
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_EXPRESSION]);
@@ -1124,3 +1306,46 @@ gtk_drop_down_get_show_arrow (GtkDropDown *self)
 
   return self->show_arrow;
 }
+
+/**
+ * gtk_drop_down_set_search_match_mode: (attributes org.gtk.Method.set_property=search-match-mode)
+ * @self: a `GtkDropDown`
+ * @search_match_mode: the new match mode
+ *
+ * Sets the match mode for the search filter.
+ *
+ * Since: 4.12
+ */
+void
+gtk_drop_down_set_search_match_mode (GtkDropDown *self,
+                                     GtkStringFilterMatchMode search_match_mode)
+{
+  g_return_if_fail (GTK_IS_DROP_DOWN (self));
+
+  if (self->search_match_mode == search_match_mode)
+    return;
+
+  self->search_match_mode = search_match_mode;
+
+  update_filter (self);
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SEARCH_MATCH_MODE]);
+}
+
+/**
+ * gtk_drop_down_get_search_match_mode: (attributes org.gtk.Method.get_property=search-match-mode)
+ * @self: a `GtkDropDown`
+ *
+ * Returns the match mode that the search filter is using.
+ *
+ * Returns: the match mode of the search filter
+ *
+ * Since: 4.12
+ */
+GtkStringFilterMatchMode
+gtk_drop_down_get_search_match_mode (GtkDropDown *self)
+{
+  g_return_val_if_fail (GTK_IS_DROP_DOWN (self), GTK_STRING_FILTER_MATCH_MODE_PREFIX);
+
+  return self->search_match_mode;
+}
