@@ -68,15 +68,17 @@
  */
 
 #include "config.h"
-#include <string.h>
-
-#include <gdk-pixbuf/gdk-pixbuf.h>
 
 #include "gtkfilefilterprivate.h"
+
 #include "gtkbuildable.h"
 #include "gtkbuilderprivate.h"
-#include "gtkprivate.h"
 #include "gtkfilter.h"
+#include "gtkprivate.h"
+
+#include <glib/gi18n-lib.h>
+#include <gdk-pixbuf/gdk-pixbuf.h>
+#include <string.h>
 
 typedef struct _GtkFileFilterClass GtkFileFilterClass;
 typedef struct _FilterRule FilterRule;
@@ -120,6 +122,9 @@ struct _FilterRule
 enum {
   PROP_0,
   PROP_NAME,
+  PROP_PATTERNS,
+  PROP_MIME_TYPES,
+  PROP_SUFFIXES,
   NUM_PROPERTIES
 };
 
@@ -149,12 +154,35 @@ gtk_file_filter_set_property (GObject      *object,
                               GParamSpec   *pspec)
 {
   GtkFileFilter *filter = GTK_FILE_FILTER (object);
+  const char * const *strv;
 
   switch (prop_id)
     {
     case PROP_NAME:
       gtk_file_filter_set_name (filter, g_value_get_string (value));
       break;
+
+    case PROP_PATTERNS:
+      strv = (const char * const *) g_value_get_boxed (value);
+      if (strv)
+        for (int i = 0; strv[i]; i++)
+          gtk_file_filter_add_pattern (filter, strv[i]);
+      break;
+
+    case PROP_MIME_TYPES:
+      strv = (const char * const *) g_value_get_boxed (value);
+      if (strv)
+        for (int i = 0; strv[i]; i++)
+          gtk_file_filter_add_mime_type (filter, strv[i]);
+      break;
+
+    case PROP_SUFFIXES:
+      strv = (const char * const *) g_value_get_boxed (value);
+      if (strv)
+        for (int i = 0; strv[i]; i++)
+          gtk_file_filter_add_suffix (filter, strv[i]);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -174,6 +202,7 @@ gtk_file_filter_get_property (GObject    *object,
     case PROP_NAME:
       g_value_set_string (value, filter->name);
       break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -196,7 +225,7 @@ filter_rule_free (FilterRule *rule)
     default:
       break;
     }
-  g_slice_free (FilterRule, rule);
+  g_free (rule);
 }
 
 static void
@@ -238,6 +267,42 @@ gtk_file_filter_class_init (GtkFileFilterClass *class)
                            NULL,
                            G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * GtkFileFilter:patterns:
+   *
+   * The patterns that this filter matches.
+   *
+   * Since: 4.10
+   */
+  props[PROP_PATTERNS] =
+      g_param_spec_boxed ("patterns", NULL, NULL,
+                          G_TYPE_STRV,
+                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GtkFileFilter:mime-types:
+   *
+   * The MIME types that this filter matches.
+   *
+   * Since: 4.10
+   */
+  props[PROP_MIME_TYPES] =
+      g_param_spec_boxed ("mime-types", NULL, NULL,
+                          G_TYPE_STRV,
+                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GtkFileFilter:suffixes:
+   *
+   * The suffixes that this filter matches.
+   *
+   * Since: 4.10
+   */
+  props[PROP_SUFFIXES] =
+      g_param_spec_boxed ("suffixes", NULL, NULL,
+                          G_TYPE_STRV,
+                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+
   g_object_class_install_properties (gobject_class, NUM_PROPERTIES, props);
 }
 
@@ -377,7 +442,7 @@ gtk_file_filter_buildable_custom_tag_start (GtkBuildable       *buildable,
 
   if (strcmp (tagname, "mime-types") == 0)
     {
-      data = g_slice_new0 (SubParserData);
+      data = g_new0 (SubParserData, 1);
       data->string = g_string_new ("");
       data->type = PARSE_MIME_TYPES;
       data->filter = GTK_FILE_FILTER (buildable);
@@ -388,7 +453,7 @@ gtk_file_filter_buildable_custom_tag_start (GtkBuildable       *buildable,
     }
   else if (strcmp (tagname, "patterns") == 0)
     {
-      data = g_slice_new0 (SubParserData);
+      data = g_new0 (SubParserData, 1);
       data->string = g_string_new ("");
       data->type = PARSE_PATTERNS;
       data->filter = GTK_FILE_FILTER (buildable);
@@ -399,7 +464,7 @@ gtk_file_filter_buildable_custom_tag_start (GtkBuildable       *buildable,
     }
   else if (strcmp (tagname, "suffixes") == 0)
     {
-      data = g_slice_new0 (SubParserData);
+      data = g_new0 (SubParserData, 1);
       data->string = g_string_new ("");
       data->type = PARSE_SUFFIXES;
       data->filter = GTK_FILE_FILTER (buildable);
@@ -426,7 +491,7 @@ gtk_file_filter_buildable_custom_tag_end (GtkBuildable *buildable,
       SubParserData *data = (SubParserData*)user_data;
 
       g_string_free (data->string, TRUE);
-      g_slice_free (SubParserData, data);
+      g_free (data);
     }
 }
 
@@ -556,7 +621,7 @@ gtk_file_filter_add_mime_type (GtkFileFilter *filter,
   g_return_if_fail (GTK_IS_FILE_FILTER (filter));
   g_return_if_fail (mime_type != NULL);
 
-  rule = g_slice_new (FilterRule);
+  rule = g_new (FilterRule, 1);
   rule->type = FILTER_RULE_MIME_TYPE;
   rule->u.content_types = g_new0 (char *, 2);
   rule->u.content_types[0] = g_content_type_from_mime_type (mime_type);
@@ -585,7 +650,7 @@ gtk_file_filter_add_pattern (GtkFileFilter *filter,
   g_return_if_fail (GTK_IS_FILE_FILTER (filter));
   g_return_if_fail (pattern != NULL);
 
-  rule = g_slice_new (FilterRule);
+  rule = g_new (FilterRule, 1);
   rule->type = FILTER_RULE_PATTERN;
   rule->u.pattern = g_strdup (pattern);
 
@@ -617,7 +682,7 @@ gtk_file_filter_add_suffix (GtkFileFilter *filter,
   g_return_if_fail (GTK_IS_FILE_FILTER (filter));
   g_return_if_fail (suffix != NULL);
 
-  rule = g_slice_new (FilterRule);
+  rule = g_new (FilterRule, 1);
   rule->type = FILTER_RULE_SUFFIX;
   rule->u.pattern = g_strconcat ("*.", suffix, NULL);
 
@@ -644,7 +709,7 @@ gtk_file_filter_add_pixbuf_formats (GtkFileFilter *filter)
 
   g_return_if_fail (GTK_IS_FILE_FILTER (filter));
 
-  rule = g_slice_new (FilterRule);
+  rule = g_new (FilterRule, 1);
   rule->type = FILTER_RULE_PIXBUF_FORMATS;
 
   array = g_ptr_array_new ();
@@ -658,9 +723,9 @@ gtk_file_filter_add_pixbuf_formats (GtkFileFilter *filter)
       mime_types = gdk_pixbuf_format_get_mime_types (l->data);
 
       for (i = 0; mime_types[i] != NULL; i++)
-        {
-          g_ptr_array_add (array, g_content_type_from_mime_type (mime_types[i]));
-        }
+        g_ptr_array_add (array, g_content_type_from_mime_type (mime_types[i]));
+
+      g_strfreev (mime_types);
     }
   g_slist_free (formats);
 
@@ -693,8 +758,6 @@ gtk_file_filter_get_attributes (GtkFileFilter *filter)
 
 #ifdef GDK_WINDOWING_MACOS
 
-#import <Foundation/Foundation.h>
-
 NSArray * _gtk_file_filter_get_as_pattern_nsstrings (GtkFileFilter *filter)
 {
   NSMutableArray *array = [[NSMutableArray alloc] init];
@@ -708,9 +771,8 @@ NSArray * _gtk_file_filter_get_as_pattern_nsstrings (GtkFileFilter *filter)
         {
         case FILTER_RULE_MIME_TYPE:
           {
-            // convert mime-types to UTI
-            NSString *mime_type_nsstring = [NSString stringWithUTF8String: rule->u.content_types[0]];
-            NSString *uti_nsstring = (NSString *) UTTypeCreatePreferredIdentifierForTag (kUTTagClassMIMEType, (CFStringRef) mime_type_nsstring, NULL);
+            // GContentType from GIO use UTI on macOS since glib version 2.51
+            NSString *uti_nsstring = [NSString stringWithUTF8String: rule->u.content_types[0]];
             if (uti_nsstring == NULL)
               {
                 [array release];
@@ -924,6 +986,13 @@ gtk_file_filter_to_gvariant (GtkFileFilter *filter)
 {
   GVariantBuilder builder;
   GSList *l;
+  GVariant *result;
+  char *name;
+
+  if (filter->name)
+    name = g_strdup (filter->name);
+  else
+    name = NULL;
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(us)"));
   for (l = filter->rules; l; l = l->next)
@@ -935,6 +1004,8 @@ gtk_file_filter_to_gvariant (GtkFileFilter *filter)
         {
         case FILTER_RULE_PATTERN:
           g_variant_builder_add (&builder, "(us)", 0, rule->u.pattern);
+          if (name == NULL)
+            name = g_strdup (rule->u.pattern);
           break;
 
         case FILTER_RULE_SUFFIX:
@@ -945,13 +1016,24 @@ gtk_file_filter_to_gvariant (GtkFileFilter *filter)
             char *pattern = _gtk_make_ci_glob_pattern (rule->u.pattern);
             g_variant_builder_add (&builder, "(us)", 0, pattern);
             g_free (pattern);
+            if (name == NULL)
+              name = g_strdup (rule->u.pattern);
           }
           break;
 
         case FILTER_RULE_MIME_TYPE:
         case FILTER_RULE_PIXBUF_FORMATS:
           for (i = 0; rule->u.content_types[i]; i++)
-            g_variant_builder_add (&builder, "(us)", 1, rule->u.content_types[i]);
+            {
+              g_variant_builder_add (&builder, "(us)", 1, rule->u.content_types[i]);
+              if (name == NULL)
+                {
+                  if (rule->type == FILTER_RULE_PIXBUF_FORMATS)
+                    name = g_strdup (_("Image"));
+                  else
+                    name = g_content_type_get_description (rule->u.content_types[i]);
+                }
+            }
           break;
 
         default:
@@ -959,7 +1041,14 @@ gtk_file_filter_to_gvariant (GtkFileFilter *filter)
         }
     }
 
-  return g_variant_new ("(s@a(us))", filter->name, g_variant_builder_end (&builder));
+  if (name == NULL)
+    name = g_strdup (_("Unspecified"));
+
+  result = g_variant_new ("(s@a(us))", name, g_variant_builder_end (&builder));
+
+  g_free (name);
+
+  return result;
 }
 
 /**
