@@ -27,8 +27,6 @@
 #include "gtkeditable.h"
 #include "gtkeventcontrollerkey.h"
 #include "gtkgestureclick.h"
-#include "gtkbox.h"
-#include "gtkimage.h"
 #include <glib/gi18n-lib.h>
 #include "gtkmarshalers.h"
 #include "gtkpasswordentrybuffer.h"
@@ -123,13 +121,10 @@ caps_lock_state_changed (GdkDevice  *device,
 {
   GtkPasswordEntry *entry = GTK_PASSWORD_ENTRY (widget);
 
-  if (gtk_editable_get_editable (GTK_EDITABLE (entry)) &&
-      gtk_widget_has_focus (entry->entry) &&
-      !gtk_text_get_visibility (GTK_TEXT (entry->entry)) &&
-      gdk_device_get_caps_lock_state (device))
-    gtk_widget_show (entry->icon);
-  else
-    gtk_widget_hide (entry->icon);
+  gtk_widget_set_visible (entry->icon, gtk_editable_get_editable (GTK_EDITABLE (entry)) &&
+                                       gtk_widget_has_focus (entry->entry) &&
+                                       !gtk_text_get_visibility (GTK_TEXT (entry->entry)) &&
+                                       gdk_device_get_caps_lock_state (device));
 }
 
 static void
@@ -213,7 +208,10 @@ gtk_password_entry_init (GtkPasswordEntry *entry)
   g_signal_connect_swapped (entry->entry, "notify::has-focus", G_CALLBACK (focus_changed), entry);
   g_signal_connect_swapped (entry->entry, "activate", G_CALLBACK (activate_cb), entry);
 
-  entry->icon = gtk_image_new_from_icon_name ("caps-lock-symbolic");
+  entry->icon = g_object_new (GTK_TYPE_IMAGE,
+                              "icon-name", "caps-lock-symbolic",
+                              "accessible-role", GTK_ACCESSIBLE_ROLE_ALERT,
+                              NULL);
   gtk_widget_set_tooltip_text (entry->icon, _("Caps Lock is on"));
   gtk_widget_add_css_class (entry->icon, "caps-lock-indicator");
   gtk_widget_set_cursor (entry->icon, gtk_widget_get_cursor (entry->entry));
@@ -231,6 +229,10 @@ gtk_password_entry_init (GtkPasswordEntry *entry)
 
   /* Transfer ownership to the GtkText widget */
   g_object_unref (buffer);
+
+  gtk_accessible_update_property (GTK_ACCESSIBLE (entry),
+                                  GTK_ACCESSIBLE_PROPERTY_HAS_POPUP, TRUE,
+                                  -1);
 }
 
 static void
@@ -253,14 +255,25 @@ gtk_password_entry_realize (GtkWidget *widget)
     }
 }
 
+static void
+gtk_password_entry_unrealize (GtkWidget *widget)
+{
+  GtkPasswordEntry *entry = GTK_PASSWORD_ENTRY (widget);
+
+  if (entry->keyboard)
+    {
+      g_signal_handlers_disconnect_by_func (entry->keyboard, caps_lock_state_changed, entry);
+      entry->keyboard = NULL;
+    }
+
+  GTK_WIDGET_CLASS (gtk_password_entry_parent_class)->unrealize (widget);
+}
+
 static void
 gtk_password_entry_dispose (GObject *object)
 {
   GtkPasswordEntry *entry = GTK_PASSWORD_ENTRY (object);
 
-  if (entry->keyboard)
-    g_signal_handlers_disconnect_by_func (entry->keyboard, caps_lock_state_changed, entry);
-
   if (entry->entry)
     gtk_editable_finish_delegate (GTK_EDITABLE (entry));
 
@@ -451,6 +464,7 @@ gtk_password_entry_class_init (GtkPasswordEntryClass *klass)
   object_class->set_property = gtk_password_entry_set_property;
 
   widget_class->realize = gtk_password_entry_realize;
+  widget_class->unrealize = gtk_password_entry_unrealize;
   widget_class->measure = gtk_password_entry_measure;
   widget_class->size_allocate = gtk_password_entry_size_allocate;
   widget_class->mnemonic_activate = gtk_password_entry_mnemonic_activate;
@@ -541,19 +555,7 @@ static gboolean
 gtk_password_entry_accessible_get_platform_state (GtkAccessible              *self,
                                                   GtkAccessiblePlatformState  state)
 {
-  GtkPasswordEntry *entry = GTK_PASSWORD_ENTRY (self);
-
-  switch (state)
-    {
-    case GTK_ACCESSIBLE_PLATFORM_STATE_FOCUSABLE:
-      return gtk_widget_get_focusable (GTK_WIDGET (entry->entry));
-    case GTK_ACCESSIBLE_PLATFORM_STATE_FOCUSED:
-      return gtk_widget_has_focus (GTK_WIDGET (entry->entry));
-    case GTK_ACCESSIBLE_PLATFORM_STATE_ACTIVE:
-      return FALSE;
-    default:
-      g_assert_not_reached ();
-    }
+  return gtk_editable_delegate_get_accessible_platform_state (GTK_EDITABLE (self), state);
 }
 
 static void
